"""S3Service module.

This module provides a class for interacting with AWS S3, allowing for operations such as 
uploading, downloading, and managing objects within an S3 bucket.
"""
from typing import List, Optional

import boto3
from botocore.client import BaseClient
from botocore.exceptions import ClientError


class S3Service:
    """
    A helper class for interacting with AWS S3.

    Parameters
    ----------
    aws_access_key_id : str
        AWS access key ID.
    aws_secret_access_key : str
        AWS secret access key.
    region_name : Optional[str]
        AWS region name. If not provided, default region will be used.

    Methods
    -------
    upload_object(bucket: str, key: str, data: bytes) -> bool:
        Uploads a file object to the specified S3 bucket.
    get_object(bucket: str, key: str) -> Optional[bytes]:
        Retrieves a file object from the specified S3 bucket.
    delete_object(bucket: str, key: str) -> bool:
        Deletes a file object from the specified S3 bucket.
    generate_presigned_url(bucket: str, key: str, expiration_time: int = 3600) -> str:
        Generates a presigned URL for the specified S3 object.
    does_key_exist(bucket: str, key: str) -> bool:
        Checks if a key exists in the specified S3 bucket.
    get_file_size(bucket: str, key: str) -> Optional[int]:
        Retrieves the size of a file object in the specified S3 bucket.
    copy(source_bucket: str, source_key: str, target_bucket: str, target_key: str) -> bool:
        Copies a file object from one bucket/key to another.
    get_all_object_names(bucket: str, prefix: str) -> List[str]:
        Retrieves all object names under a specific prefix (folder path) in the S3 bucket.
        Returns an empty list if no objects are found.
    get_all_object_urls(bucket: str, prefix: str) -> List[str]:
        Retrieves the URLs of all objects under a specific prefix in the S3 bucket.
        Returns an empty list if no objects are found.
    """

    def __init__(
            self,
            aws_access_key_id: str,
            aws_secret_access_key: str,
            region_name: Optional[str] = None) -> None:
        """
        Initializes the S3Service with AWS credentials.

        Parameters
        ----------
        aws_access_key_id : str
            AWS access key ID.
        aws_secret_access_key : str
            AWS secret access key.
        region_name : Optional[str]
            AWS region name. If not provided, default region will be used.
        """
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.region_name = region_name
        self.client = self._create_s3_client()

    def _create_s3_client(self) -> BaseClient:
        """
        Creates an S3 client using the provided AWS credentials.

        Returns
        -------
        BaseClient
            A boto3 S3 client instance.
        """
        return boto3.client('s3',
                            aws_access_key_id=self.aws_access_key_id,
                            aws_secret_access_key=self.aws_secret_access_key,
                            region_name=self.region_name)

    def _handle_info_response(self, response: dict) -> bool:
        """
        Handles S3 responses that provide informational status.

        Parameters
        ----------
        response : dict
            The response dictionary from an S3 operation.

        Returns
        -------
        bool
            True if the operation was successful, False otherwise.
        """
        status_code = response.get(
            "ResponseMetadata", {}).get("HTTPStatusCode", 0)
        return status_code in [200, 204]

    def upload_object(self, bucket: str, key: str, data: bytes) -> bool:
        """
        Uploads a file object to the specified S3 bucket.

        Parameters
        ----------
        bucket : str
            The name of the S3 bucket.
        key : str
            The object key in the S3 bucket.
        data : bytes
            The data to upload.

        Returns
        -------
        bool
            True if the upload was successful, False otherwise.
        """
        try:
            response = self.client.put_object(
                Bucket=bucket, Key=key, Body=data)
            return self._handle_info_response(response)
        except ClientError as e:
            print(f"Error uploading file: {e}")
            return False

    def get_object(self, bucket: str, key: str) -> Optional[bytes]:
        """
        Retrieves a file object from the specified S3 bucket.

        Parameters
        ----------
        bucket : str
            The name of the S3 bucket.
        key : str
            The object key in the S3 bucket.

        Returns
        -------
        Optional[bytes]
            The file data if successful, None otherwise.
        """
        try:
            response = self.client.get_object(Bucket=bucket, Key=key)
            return response['Body'].read()
        except ClientError as e:
            print(f"Error getting file: {e}")
            return None

    def delete_object(self, bucket: str, key: str) -> bool:
        """
        Deletes a file object from the specified S3 bucket.

        Parameters
        ----------
        bucket : str
            The name of the S3 bucket.
        key : str
            The object key in the S3 bucket.

        Returns
        -------
        bool
            True if the deletion was successful, False otherwise.
        """
        try:
            response = self.client.delete_object(Bucket=bucket, Key=key)
            return self._handle_info_response(response)
        except ClientError as e:
            print(f"Error deleting file: {e}")
            return False

    def generate_presigned_url(
        self, bucket: str, key: str, expiration_time: int = 3600
    ) -> str:
        """
        Generates a presigned URL for the specified S3 object.

        Parameters
        ----------
        bucket : str
            The name of the S3 bucket.
        key : str
            The object key in the S3 bucket.
        expiration_time : int, optional
            Time in seconds for the presigned URL to remain valid, by default 3600.

        Returns
        -------
        str
            The presigned URL.
        """
        try:
            url = self.client.generate_presigned_url(
                'get_object',
                Params={'Bucket': bucket, 'Key': key},
                ExpiresIn=expiration_time,
            )
            return url
        except ClientError as e:
            print(f"Error generating presigned URL: {e}")
            return ""

    def does_key_exist(self, bucket: str, key: str) -> bool:
        """
        Checks if a key exists in the specified S3 bucket.

        Parameters
        ----------
        bucket : str
            The name of the S3 bucket.
        key : str
            The object key to check.

        Returns
        -------
        bool
            True if the key exists, False otherwise.
        """
        try:
            self.client.head_object(Bucket=bucket, Key=key)
            return True
        except ClientError as e:
            if e.response['Error']['Code'] == '404':
                return False
            print(f"Error checking if key exists: {e}")
            return False

    def get_file_size(self, bucket: str, key: str) -> Optional[int]:
        """
        Retrieves the size of a file object in the specified S3 bucket.

        Parameters
        ----------
        bucket : str
            The name of the S3 bucket.
        key : str
            The object key in the S3 bucket.

        Returns
        -------
        Optional[int]
            The size of the file in bytes if successful, None otherwise.
        """
        try:
            response = self.client.head_object(Bucket=bucket, Key=key)
            return response['ContentLength']
        except ClientError as e:
            print(f"Error getting file size: {e}")
            return None

    def copy(
        self, source_bucket: str, source_key: str, target_bucket: str, target_key: str
    ) -> bool:
        """
        Copies a file object from one bucket/key to another.

        Parameters
        ----------
        source_bucket : str
            The name of the source S3 bucket.
        source_key : str
            The object key in the source S3 bucket.
        target_bucket : str
            The name of the target S3 bucket.
        target_key : str
            The object key in the target S3 bucket.

        Returns
        -------
        bool
            True if the copy was successful, False otherwise.
        """
        try:
            copy_source = {'Bucket': source_bucket, 'Key': source_key}
            self.client.copy(copy_source, target_bucket, target_key)
            return True
        except ClientError as e:
            print(f"Error copying file: {e}")
            return False

    def get_all_object_names(self, bucket: str, prefix: str) -> List[str]:
        """
        Retrieves all object names under a specific prefix (folder path) in the S3 bucket.

        This method paginates through the S3 bucket and collects the keys of all objects
        that match the specified prefix.

        Parameters
        ----------
        bucket : str
            The name of the S3 bucket.
        prefix : str
            The prefix (folder path) to list objects from. Can be empty to list all objects.

        Returns
        -------
        List[str]
            A list of object keys under the specified prefix. If an error occurs, 
            an empty list is returned.
        """
        object_names = []
        paginator = self.client.get_paginator('list_objects_v2')
        try:
            for page in paginator.paginate(Bucket=bucket, Prefix=prefix):
                contents = page.get('Contents', [])
                object_names.extend([obj['Key'] for obj in contents])
            return object_names
        except ClientError as e:
            print(f"Error listing objects: {e}")
            return []

    def get_all_object_urls(self, bucket: str, prefix: str) -> List[str]:
        """
        Retrieves the URLs of all objects under a specific prefix in the S3 bucket.

        Parameters
        ----------
        bucket : str
            The name of the S3 bucket.
        prefix : str
            The prefix (folder path) to list objects from.

        Returns
        -------
        List[str]
            A list of object URLs.
        """
        object_urls: List[str] = []
        region: str = self.client.meta.region_name or 'us-east-1'
        object_names: List[str] = self.get_all_object_names(bucket, prefix)
        for key in object_names:
            url = f"https://{bucket}.s3.{region}.amazonaws.com/{key}"
            object_urls.append(url)
        return object_urls
