from collections import defaultdict

from setuptools import find_packages, setup


def parse_requirements_file(path, allowed_extras: set = None, include_all_extra: bool = True):
    requirements = []
    extras = defaultdict(list)
    with open(path) as requirements_file:
        import re

        def fix_url_dependencies(req: str) -> str:
            """Pip and setuptools disagree about how URL dependencies should be handled."""
            m = re.match(
                r"^(git\+)?(https|ssh)://(git@)?github\.com/([\w-]+)/(?P<name>[\w-]+)\.git", req
            )
            if m is None:
                return req
            else:
                return f"{m.group('name')} @ {req}"

        for line in requirements_file:
            line = line.strip()
            if line.startswith("#") or len(line) <= 0:
                continue
            req, *needed_by = line.split("# needed by:")
            req = fix_url_dependencies(req.strip())
            if needed_by:
                for extra in needed_by[0].strip().split(","):
                    extra = extra.strip()
                    if allowed_extras is not None and extra not in allowed_extras:
                        raise ValueError(f"invalid extra '{extra}' in {path}")
                    extras[extra].append(req)
                if include_all_extra and req not in extras["all"]:
                    extras["all"].append(req)
            else:
                requirements.append(req)
    return requirements, extras


integrations = {"checklist"}
install_requirements, extras = parse_requirements_file(
    "requirements.txt", allowed_extras=integrations
)

# version.py defines the VERSION and VERSION_SHORT variables.
# We use exec here so we don't import imutils whilst setting up.
VERSION = {}  # type: ignore
with open("s3service/version.py", "r") as version_file:
    exec(version_file.read(), VERSION)

setup(
    name="s3service",
    version=VERSION["VERSION"],
    description="A library for managing AWS S3 operations.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Developers",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
    ],
    keywords="AWS, S3, Python, cloud storage",
    url="https://github.com/myth-ai/s3service",
    author="Myth-AI Research for Generative Artificial Intelligence",
    author_email="bugra@myth-ai.com",
    license="Apache",
    packages=find_packages(
        exclude=[
            "*.tests",
            "*.tests.*",
            "tests.*",
            "tests",
            "test_fixtures",
            "test_fixtures.*",
            "benchmarks",
            "benchmarks.*",
        ]
    ),
    install_requires=install_requirements,
    extras_require=extras,
    include_package_data=True,
    python_requires=">=3.7",
    zip_safe=False,
)
